module.exports = (env) ->

  Promise = env.require 'bluebird'
  assert = env.require 'cassert'
  _ = env.require 'lodash'

  M = env.matcher

  # ping is used for IP address based checks
  ping = require 'ping'

  # exec is used for bluetooth based checks to run hcitool
  exec = require('child_process').exec

  # ###PimaticNetCheck class
  class PimaticNetCheck extends env.plugins.Plugin
    init: (app, @framework, @config) =>
      env.logger.debug('config', @config)

      switch @framework.config.settings.locale
        when "de" then deviceConfigDef = require("./device-config-schema-de")
        else deviceConfigDef = require("./device-config-schema")

      @framework.deviceManager.registerDeviceClass('NetCheckDevice', {
        configDef: deviceConfigDef.NetCheckDevice,
        createCallback: (deviceConfig, lastState) =>
          device = new NetCheckDevice(deviceConfig, lastState, @framework)
          return device
      })

      @framework.ruleManager.addActionProvider new NetCheckActionProvider(@framework, @config)

  # ###NetCheckDevice class
  class NetCheckDevice extends env.devices.PresenceSensor

    actions:
      check:
        description: "executes a check"


    constructor: (@config, lastState, @framework) ->
      env.logger.debug('NetCheckDevice', @config, lastState)

      @id = @config.id
      @name = @config.name
      @_presence = lastState?.presence?.value or false

      @needAll = if @config.needAll? then @config.needAll else true

      @checkInterval = @config.checkInterval or 0

      @bluetoothDevice = @config.bluetoothDevice or 'hci0'

      @bluetoothSudo = if @config.bluetoothUseSudo then 'sudo ' else ''

      @pingTimeout = @config.pingTimeout or 2

      assert @config.networkIpAddresses.length > 0 or @config.bluetoothMacAddresses.length > 0, "You need to define at least one IP or MAC address"

      # check the configured MAC addresses
      for mac in @config.bluetoothMacAddresses
        assert mac.match(/^[0-9a-fA-F]{2}\:[0-9a-fA-F]{2}\:[0-9a-fA-F]{2}\:[0-9a-fA-F]{2}\:[0-9a-fA-F]{2}\:[0-9a-fA-F]{2}$/), "Invalid MAC address"

      # checkInterval should not be less than one second
      assert @checkInterval is 0 or @checkInterval > 999, "The checkInterval must be zero or grater than 999"

      # start the interval for automatic checks
      if @checkInterval > 0
        env.logger.debug "#{@id} starting automatic checking every #{@checkInterval} ms"
        @checkIntervalInterval = setInterval ( => @check() ), @checkInterval
      else
        @checkIntervalInterval = null

      super()

    destroy: ->
      clearInterval @checkIntervalInterval
      super()

    # ####check()
    # Function to start the check
    check: ->
      env.logger.debug "#{@id} check started, needAll: #{@needAll}"

      promise = Promise.resolve()

      okCount = 0
      errorCount = 0

      # check the network IP addresses
      for ip in @config.networkIpAddresses
        do (@id, ip, @needAll, @pingTimeout) ->
          promise = promise.then( =>
            # no need to do other checks if all ips needed and already one is not found
            # or no need to do other checks if not all ips needed and already one is found
            if (@needAll and errorCount > 0) or (not @needAll and okCount > 0)
              return

            env.logger.debug "#{@id} checking network #{ip}"
            return ping.promise.probe(ip, timeout: @pingTimeout)
            .then( (pingResponse) =>
              if pingResponse.alive
                # ping ok
                env.logger.debug "#{@id} network #{ip} present"
                okCount++;
              else
                # ping not ok
                env.logger.debug "#{@id} network #{ip} absent"
                errorCount++;
            )
          )

      # check the bluetooth MAC addresses
      for mac in @config.bluetoothMacAddresses
        do (@id, mac, @needAll, @bluetoothDevice, @bluetoothSudo, @pingTimeout) ->
          promise = promise.then( =>
            # no need to do other checks if all ips needed and already one is not found
            # or no need to do other checks if not all ips needed and already one is found
            if (@needAll and errorCount > 0) or (not @needAll and okCount > 0)
              return

            env.logger.debug "#{@id} checking bluetooth #{mac}"
            # new promise for running hcitool
            return new Promise( (resolve, reject) =>
              exec "#{@bluetoothSudo}l2ping -i #{@bluetoothDevice} -c 1 -t #{@pingTimeout} #{mac}", (error, stdout, stderr) =>
                if error
                  if error.code is 1
                    # device absent
                    env.logger.debug "#{@id} bluetooth #{mac} absent"
                    errorCount++
                  else
                    # other error
                    env.logger.warn "#{@id} error running 'l2ping':", error
                    resolve()
                    return
                else
                  # device present
                  # even if no ping response...
                  env.logger.debug "#{@id} bluetooth #{mac} present"
                  okCount++

                # resolve the l2ping-promise
                resolve()
            )
          )

      # if all checks are done
      promise.then( =>

        # evaluate the results
        if (@needAll && errorCount is 0) or ( not @needAll and okCount > 0)
          # needed all devices and no error
          # or not needed all and min one ok
          # check succeeded
          env.logger.debug "#{@id} netcheck done and succeeded"

          # set as present
          @_setPresence true
        else
          # check failed
          env.logger.debug "#{@id} netcheck done but failed"

          # set as absent
          @_setPresence false
      )

      # return the promise chain
      return promise

  # ###NetCheckActionProvider class
  class NetCheckActionProvider extends env.actions.ActionProvider
    constructor: (@framework, @config) ->

    parseAction: (input, context) =>
      rulePresentId = null
      ruleAbsentId = null
      deviceId = null

      # get all rules from the ruleManager
      rules = @framework.ruleManager.getRules()
      rulesWithId = _(rules).map( (r) => [r, r.id] ).value()
      # fix for no loaded rules on startup
      if rulesWithId.length is 0
        rulesWithId = _.map(@framework.config.rules, (r) => [r, r.id] )

      # get all NetCheckDevices
      netCheckDevices = _(@framework.deviceManager.devices).values()
        .filter((device) => device instanceof NetCheckDevice).value()

      # rule syntax:
      #  check <netcheck-deviceID> [if present exec rule <ruleID>] [if absent exec rule <ruleID>]
      m = M(input, context)
        .match("check ")
        .matchDevice(netCheckDevices, (m, d) =>
          deviceId = d.id
        )
        .optional( (m) =>
          next = m
          m.match(" if present exec rule ")
          .match(rulesWithId, (m,r) =>
            next = m
            rulePresentId = r.id
          )
          return next
        )
        .optional( (m) =>
          next = m
          m.match(" if absent exec rule ")
          .match(rulesWithId, (m,r) =>
            next = m
            ruleAbsentId = r.id
          )
          return next
        )

      if m.hadMatch()
        match = m.getFullMatch()
        return {
          token: match
          nextInput: input.substring(match.length)
          actionHandler: new NetCheckActionHandler(@framework, deviceId, rulePresentId, ruleAbsentId)
        }
      else
        return null

  # ###NetCheckActionHandler class
  class NetCheckActionHandler extends env.actions.ActionHandler
    constructor: (@framework, @deviceId, @rulePresentId, @ruleAbsentId) ->

    executeAction: (simulate) =>
      # get the device
      device = @framework.deviceManager.getDeviceById @deviceId
      assert device instanceof NetCheckDevice

      # get the rules
      rulePresent = @framework.ruleManager.getRuleById @rulePresentId
      ruleAbsent = @framework.ruleManager.getRuleById @ruleAbsentId

      if simulate
        return Promise.resolve __("#{device.id} would execute a check")
      else
        # run the check
        device.check()
        .then( =>
          # check presence as result of the check
          if device._presence
            env.logger.debug "#{device.id} check returned present"
            if rulePresent? and rulePresent.active
              env.logger.debug "#{device.id} executing rule present #{@rulePresentId}"
              @framework.ruleManager._executeRuleActionsAndLogResult rulePresent
              return Promise.resolve __("executing rule present")

          else
            env.logger.debug "#{device.id} check returned absent"
            if ruleAbsent? and ruleAbsent.active
              env.logger.debug "#{device.id} executing rule absent #{@ruleAbsentId}"
              @framework.ruleManager._executeRuleActionsAndLogResult ruleAbsent
              return Promise.resolve __("executing rule absent")

          return Promise.resolve __("executing no rule")
        )

  # Create a instance of the plugin
  pimaticNetCheck = new PimaticNetCheck

  # ###Finally
  # return it to the framework.
  return pimaticNetCheck
